--[[
pge.http

A more flexible HTTP module using pge.net.socket.

Provides the following functions:
   string = urlencode(string)
   string = urldecode(string)
      Respectively, encode and decode URL escape codes (%xx).

   scheme, host, port, path = parseurl(url)
      Tries to parse a URL; on success returns each segment.
      If no port was specified, tries to use a reasonable default.

   boolean, result = getstring(url)
      Retrieves a URL using HTTP GET, returning the contents as a string.
      On success, returns 'true' and the contents.
      On failure, returns 'false', an error code, and an error message.

   boolean, number, string = getfile(url, filename)
      Retrieves a URL using HTTP GET, saving the contents to a file.
      Returns 'true' or 'false', a result code, and a result message.

   boolean, result = postform(url, data, ...)
      Retrieves a URL using HTTP POST, returning the contents as a string.
      See getstring for return values.

Additionally, it can be customized by setting these values:
   useragent
      (string) Overrides the default "User-Agent" header value.
   keepalive
      (boolean) If 'true', each request will attempt to keep the connection
      open for later requests to the same host.
]]

--------------------------------------------------------------------------------
--[[
local pge = pge
local type = type
local setmetatable = setmetatable
local getmetatable = getmetatable
local tonumber = tonumber
local tostring = tostring

module('pge.http', package.seeall)
]]

do
pge.http = {}
local _M = pge.http

local BLKSIZE = 4096

local DEFAULT = {
   useragent = 'Mozilla/4.0 (PSP (PlayStation Portable); 2.00)',
   keepalive = true
}

local SETTING = setmetatable({}, { __index = DEFAULT })

require "logging"
local log = logging.start()

log:write("============================================================\n")
log:write(("Begin log: %s\n"):format(os.date()))
log:write("------------------------------------------------------------\n")

--------------------------------------------------------------------------------

function _M.urlencode(s)
   return s:gsub('[^%w._-]', function(t) return ('%%%02X'):format(t:byte()) end)
end
function _M.urldecode(s)
   return s:gsub('%%(%x%x)', function(t) return string.char(tonumber(t,16)) end)
end

--------------------------------------------------------------------------------

-- List of common service ports. This library only supports HTTP, but
-- the parseurl function might be useful elsewhere.
local service_ports = {
   http = 80,
   https = 443,
   ftp = 21,
   pop3 = 110,
   smtp = 25,
   imap = 143,
}
function _M.parseurl(url)
   local scheme, host, port, path =
      url:match('^(%a[%w.+-]*)://([^/:])(:[^/]*)(.*)$')

   if port then
      port = port:sub(2)
   else
      scheme, host, path = url:match('^(%a[%w.+-]*)://([^/]+)(.*)$')
   end

   if scheme then
      scheme = scheme:lower()

      port = tonumber(port) or service_ports[scheme] or 80

      --print(scheme,host,port,path)

      return scheme, host, port, path
   end
end

--------------------------------------------------------------------------------

-- A simple connection manager to help implement keep-alive.

local conns = {}
local connMRU = {}
local MAXCONNS = 8

-- Find a socket in the MRU list.
local function sock_find(sck)
   for i, v in ipairs(connMRU) do
      if v == sck then return i end
   end
end

-- Close a given socket.
local function sock_close(sck)
   if type(sck) == 'number' then
      sck = connMRU[sck]
   end

   local id = conns[sck]
   if id then
      table.remove(connMRU, sock_find(sck))

      print(('closing connection to %s'):format(id))

      conns[sck] = nil
      conns[id ] = nil
      sck:close()
   end
end

-- Get a socket connection to the given ip and port.
local function sock_connect(ip, port)
   if ip and port then
      local sck, id

      id = ('%s:%d'):format(ip, port)

      if conns[id] then
         -- Reuse an open connection.

         print(('reusing connection to %s'):format(id))

         sck = conns[id]
         table.remove(connMRU, sock_find(sck))
      else
         -- Open a new connection.

         print(('opening connection to %s'):format(id))

         sck = pge.net.socket.create()

         if sck and sck:connect(ip, port) then
            if #connMRU > MAXCONNS then
               -- close the LRU open socket
               sock_close(1)
            end
            conns[sck] = id
            conns[id] = sck
         end
      end

      if sck then
         connMRU[#connMRU+1] = sck
         return sck
      --else
      --   error(('error connecting to %s'):format(id))
      end
   end
end

--------------------------------------------------------------------------------

-- Address resolution cache.
local resolve = setmetatable({}, {
   __index = function(t, k)
      if k:match('^%d+[.]%d+[.]%d+[.]%d+$') then
         return k
      else
         k = tostring(k):lower()

         print(('resolving "%s"'):format(k))

         local r = pge.net.resolve(k)

         if r then
            log:write(("host '%s' resolved to %s\n"):format(k, r))
            t[k] = r
            return t[k]
         else
            log:write(("host '%s' did not resolve\n"):format(k))
         --   error('hostname not found')
         end
      end
   end,
   })

--------------------------------------------------------------------------------

-- Sends a request to the given URL, opening a new connection if necessary.

local function sendReq(verb, url, headers, content)
   if not pge.net.isconnected() then return end

   print(('retrieving "%s"'):format(url))

   local scheme, host, port, path = _M.parseurl(url)

   -- can only handle HTTP requests
   if scheme ~= 'http' then
      print(('invalid scheme "%s"'):format(scheme))
      return
   end

   if host then
      if #path == 0 then path = '/' end

      --print('connecting to: '..host)

      local sck, msg = sock_connect(resolve[host], port)

      if not sck then return end

      --print('connected, requesting '..path)

      log:write("------------------------------------------------------------\n")

      local rqln = ('%s %s HTTP/1.1\r\n'):format(verb, path:gsub(' ', '%%20'))
      log:write(rqln)
      sck:send(rqln)

      local hdrs = {}
      for k, v in pairs(headers or {}) do
         hdrs[tostring(k):lower()] = v
      end

      --print('sending headers')

      hdrs['host'] = host
      --hdrs['te'] = 'identity'

      if SETTING.keepalive then
         --hdrs['connection'] = 'Keep-Alive'
      end

      if content then
         content = tostring(content)
         hdrs['content-length'] = #content
      else
         hdrs['content-length'] = nil
      end

      if not hdrs['user-agent'] then
         hdrs['user-agent'] = SETTING.useragent
      end

      for k, v in pairs(hdrs) do
         local hdln = ('%s: %s\r\n'):format(tostring(k), tostring(v))
         log:write(hdln)
         sck:send(hdln)
      end

      sck:send('\r\n')

      if content then
         --print('sending content')
         sck:send(content)
      end

      return sck
   end
end

--------------------------------------------------------------------------------

-- Basic data handlers for readRsp:
--    file_handler   - writes data to a file
--    string_handler - returns data as a string

-- Later versions might include a way to add custom handlers.

local file_handler = {
   __type = "file_handler",

   __init = function(t, filename)
      t.filename = filename
   end,

   open = function(t)
      if not t.fh then
         print(('opening "%s"'):format(t.filename))
         t.fh = pge.file.open(t.filename,
            PGE_FILE_WRONLY +
            PGE_FILE_CREATE +
            PGE_FILE_TRUNC)
         return t.fh and true or false
      end
   end,

   close = function(t)
      if t.fh then
         print(('closing "%s"'):format(t.filename))
         t.fh:close()
         t.fh = nil
      end
   end,

   write = function(t, data)
      if t.fh then
         t.fh:write(data)
      end
   end,
}
local string_handler = {
   __type = "string_handler",

   open = function(t)
      t.data = {}
      return true
   end,
   close = function(t)
      t.string = table.concat(t.data)
      t.data = nil
   end,
   write = function(t, data)
      t.data[#t.data+1] = data
   end,
}

local chunked_handler = {
   __init = function(t, inner)
      log:write(("chunked_handler %s wrapping %s %s"):format(
         tostring(t), inner.__type, tostring(inner)))
      t.inner = inner
      t.data = ""
      t.chunklen = 0
   end,
   open = function(t)
      log:write("starting chunked transfer\n")
      return t.inner:open()
   end,
   writechunk = function(t)
      if #t.data >= t.chunklen then
         if t.chunklen > 0 then
            t.inner:write(t.data:sub(1,t.chunklen))
         end
         t.data = t.data:sub(t.chunklen+1)
         local a,b = t.data:find("([0-9]+) *\r\n")
         if a and b then
            t.chunklen = tonumber(t.data:sub(a,b))
            log:write(("begin chunk, length = %d\n"):format(t.chunklen))
            t.data = t.data:sub(b+1)
         else
            if #data > 50 then
               data = data:sub(1,47).."..."
            end
            log:write(("no chunk length found: '%s'"):format(data))
         end
         return true
      end
   end,
   close = function(t)
      log:write("ending chunked transfer\n")
      while pge.net.isconnected() and t:writechunk() do end
      log:flush()
      return t.inner:close()
   end,
   write = function(t, data)
      log:write(("chunked_handler: got %d bytes\n"):format(#data))
      t.data = t.data .. data
      while pge.net.isconnected() and t:writechunk() do end
   end,
}

local function createHandler(h, ...)
   local o = setmetatable({}, { __index = h, __metatable = h })
   if o.__init then o:__init(...) end
   return o
end

--------------------------------------------------------------------------------

-- Read an HTTP response from a socket, using the given handler to write the
-- data returned (content only, no headers).
local function readRsp(sck, handler)
   if not sck then
      return false, 999, 'invalid socket handle'
   end

   log:write("------------------------------------------------------------\n")

   local sset = pge.net.socketset.create()
   sset:add(sck)

   local rsp = sck:receive(BLKSIZE)
   local _, eoh = rsp:find('\r\n\r\n')

   while pge.running() and pge.net.isconnected() and not eoh do
      local t
      while pge.running() and pge.net.isconnected() and sset:select() == 0 do sset:add(sck) end
      t = sck:receive(BLKSIZE)
      if #t == 0 then break end
      rsp = rsp .. t
      _, eoh = rsp:find('\r\n\r\n')
   end

   local stat = rsp:find('\r\n')
   local ver, code, msg

   if stat then
      ver, code, msg = rsp:sub(1, stat - 1):match('^(%S+) (%d+) (.*)')
      code = tonumber(code)
   else
      return false, 999, 'missing HTTP status-line'
   end

   log:write(("%s %s %s\n"):format(ver,code,msg))

   local headers = {}
   for name, value in rsp:sub(stat + 2, eoh - 2):gmatch('([%w-]+): *(.-)\r\n') do
      headers[name:lower()] = value
      log:write(("%s: %s\n"):format(name, value))
   end

   if code < 400 then
      if headers['transfer-encoding'] == 'chunked' then
         handler = createHandler(chunked_handler, handler)
      end
      if handler:open() then
         handler:write(rsp:sub(eoh + 1))

         local t = sck:receive(BLKSIZE)
         while pge.net.isconnected() and #t > 0 do
            --print(('read %d bytes'):format(#t))
            handler:write(t)
            t = sck:receive(BLKSIZE)
         end

         handler:close()
      else
         return false, 999, 'error opening handler'
      end
   end

   if (headers['connection'] == 'close') or not SETTING.keepalive then
      sock_close(sck)
   end

   return code < 400, code, msg
end

--------------------------------------------------------------------------------

function _M.getstring(url)
   local h = createHandler(string_handler)
   local ok, code, msg = readRsp(sendReq('GET', url), h)

   if ok then
      return ok, h.string
   else
      return ok, msg
   end
end

function _M.getfile(url, file)
   local h = createHandler(file_handler, file)

   local ok, code, msg = readRsp(sendReq('GET', url), h)

   return ok, msg
end

function _M.postform(url, data)
   local h = createHandler(string_handler)
   local ok, code, msg = readRsp(sendReq('POST', url, {}, data), h)

   if ok then
      return ok, h.string
   else
      return ok, msg
   end
end

--[[
-- Needs a better name.  Probably not as useful as 'getstring' anyway.
function _M.postfile(url, file, data)
   local h = createHandler(file_handler, file)
   local sck = sendReq('POST', url, {}, data)

   return readRsp(sck, h)
end
]]

setmetatable(_M, {
   __index = SETTING,
   __newindex = SETTING,
   })

end
