--------------------------------------------------------------------------------
-- Lua table extensions

do
   local __dump
   __dump = {
      ['table'] = function (tbl, r)
         r[#r+1] = '{'
         for k, v in pairs(tbl) do
            local kt, vt = type(k), type(v)
            local kf = __dump[kt]
            local vf = __dump[vt]
            if kf and vf then
               r[#r+1] = '['
               kf(k, r)
               r[#r+1] = ']='
               vf(v, r)
               if next(tbl,k) then
                  r[#r+1] = ','
               end
            end
         end
         r[#r+1] = '}'
      end,
      ['string'] = function (str, r)
         r[#r+1] = ('%q'):format(str)
      end,
      ['number'] = function (num, r)
         r[#r+1] = ('%.14g'):format(num)
      end,
      ['boolean'] = function (bool, r)
         r[#r+1] = bool and 'true' or 'false'
      end,
      ['function'] = function (func, r)
         r[#r+1] = ('loadstring(%q)'):format(string.dump(func))
      end,
   }

   -- Dumps any value (except 'userdata' or 'thread') into
   -- a string that can be restored with loadstring.
   -- (table.concat suggested by doub on #Lua - thanks!)
   function table.dump(v)
      local r = { 'return ' }
      __dump[type(v)](v, r)
      return table.concat(r)
   end
end

-- Scans a table for an item. If the item is found,
-- returns its key, else nil.
function table.find(tbl, item)
   for k, v in pairs(tbl) do
      if v == item then return k end
   end
end

-- Uses a binary search to find an item. If the item
-- is found, returns its index, else nil.
function table.search(tbl, item)
   local a, b = 1, #tbl
   if tbl[a] == item then return a end
   if tbl[b] == item then return b end
   while b - a > 1 do
      local m = floor((a + b) * 0.5)
      if tbl[m] == item then
         return m
      elseif tbl[m] < item then
         a = m
      else
         b = m
      end
   end
end

-- Copies the fields of one table to another. Returns the copy.
function table.copy(tbl, copy)
   if copy == nil then copy = {} end
   for k, v in pairs(tbl) do
      copy[k] = v
   end
   return copy
end

-- Returns a generator to iterate over a given range of items.
function table.range (t, i, j)
   i = i or 1
   j = j or #t
   --if i < 0 then i = #t + i end
   --if j < 0 then j = #t + j end
   return function()
      if i <= j and t[i] then
         local n = i
         i = i + 1
         return n, t[n]
      end
   end
end

-- Reduces a table using the given function.
-- An optional initial value may be given.
function table.reduce(tbl, fn, init)
   local rv,k,v = init
   if rv == nil then
      k,rv = next(tbl)
   end
   repeat
      k, v = next(tbl, k)
      if k then
         rv = fn(rv, v)
      end
   until k == nil
   return rv
end

-- Filters a table using the given function.
function table.filter(tbl, fn)
   local t = {}
   for k,v in pairs(tbl) do
      if fn(v) then
         t[k] = v
      end
   end
   return t
end

-- Applies the given function to each value of a table.
function table.map(tbl, fn)
   local t = {}
   for k,v in pairs(tbl) do
      t[k] = fn(v)
   end
   return t
end

